<?php

declare(strict_types=1);

namespace SimpleSAML\Module\privacyidea\Auth\Process;

use SimpleSAML\Auth\ProcessingChain;
use SimpleSAML\Auth\ProcessingFilter;
use SimpleSAML\Auth\State;
use SimpleSAML\Error\ConfigurationError;
use SimpleSAML\Logger;
use SimpleSAML\Module;
use SimpleSAML\Module\privacyidea\Auth\Utils;
use SimpleSAML\Module\saml\Error\NoPassive;
use SimpleSAML\Utils\HTTP;

/**
 * This authentication processing filter allows you to add a second step authentication against privacyIDEA.
 */
class PrivacyideaAuthProc extends ProcessingFilter
{
    /**
     * @var array This contains the authproc configuration which is set in metadata
     */
    private $authProcConfig;
    /**
     * @var PrivacyIDEA This is an object from privacyIDEA class
     */
    private $pi;

    /**
     * @param array $config   authproc configuration
     * @param mixed $reserved
     */
    public function __construct(array $config, $reserved)
    {
        assert('array' === gettype($config));
        parent::__construct($config, $reserved);
        $this->authProcConfig = $config;
        $this->pi = Utils::createPrivacyIDEAInstance($config);
        if (null === $this->pi) {
            throw new ConfigurationError('privacyIDEA: Initialization failed.');
        }
    }

    /**
     * Run the filter.
     *
     * @param array $state
     */
    public function process(&$state)
    {
        Logger::info('privacyIDEA: Auth Proc Filter - Entering process function.');
        assert('array' === gettype($state));

        // Update state before starting the authentication process
        $state['privacyidea:privacyidea'] = $this->authProcConfig;
        $state['privacyidea:privacyidea']['authenticationMethod'] = 'authprocess';
        $state['privacyidea:privacyidea:ui']['showLogout'] = $this->authProcConfig['showLogout'] ?? true;
        $state['privacyidea:privacyidea:ui']['messageOverride'] = $this->authProcConfig['messageOverride'] ?? null;

        // If set in config, allow to check the IP of the client and to control the 2FA depending on the client IP.
        // It can be used to configure that a user does not need to provide a second factor when logging in from the local network.
        if (!empty($this->authProcConfig['excludeClientIPs'])) {
            $ip = Utils::getClientIP();
            if ($this->matchIP($ip, $this->authProcConfig['excludeClientIPs'])) {
                Logger::debug('privacyIDEA: privacyIDEA is disabled because ip ' . $ip . ' is excluded.');
                ProcessingChain::resumeProcessing($state);
            }
        }

        // If set to true in config, selectively disable the privacyIDEA authentication using the entityID and/or SAML attributes.
        // The skipping will be done in self::isPrivacyIDEADisabled
        if (!empty($this->authProcConfig['checkEntityID']) && true === $this->authProcConfig['checkEntityID']) {
            $stateId = State::saveState($state, 'privacyidea:privacyidea');
            $stateId = $this->checkEntityID($this->authProcConfig, $stateId);
            $state = State::loadState($stateId, 'privacyidea:privacyidea');
        }

        // Check if privacyIDEA is disabled by configuration setting
        if (self::isPrivacyIDEADisabled($state, $this->authProcConfig)) {
            Logger::debug('privacyIDEA: privacyIDEA is disabled by a filter');
            ProcessingChain::resumeProcessing($state);
        }

        // SSO check if authentication should be skipped
        if (array_key_exists('SSO', $this->authProcConfig)
            && true === $this->authProcConfig['SSO']) {
            if (Utils::checkForValidSSO($state)) {
                Logger::debug('privacyIDEA: SSO data valid - logging in..');
                ProcessingChain::resumeProcessing($state);
            } else {
                Logger::debug('privacyIDEA: No valid SSO data found.');
            }
        }

        $username = $state['Attributes'][$this->authProcConfig['uidKey']][0];
        $stateId = State::saveState($state, 'privacyidea:privacyidea');

        if (isset($state['isPassive']) && true === $state['isPassive']) {
            throw new NoPassive('Passive multi-factor authentication not supported.');
        }

        // Check if it should be controlled that user has no tokens and a new token should be enrolled.
        if (!empty($this->authProcConfig['doEnrollToken']) && true === $this->authProcConfig['doEnrollToken']) {
            $stateId = $this->enrollToken($stateId, $username);
        }

        // Check if triggerChallenge or a call with a static pass to /validate/check should be done
        if (!empty($this->authProcConfig['doTriggerChallenge']) && true === $this->authProcConfig['doTriggerChallenge']) {
            // Call /validate/triggerchallenge with the service account from the configuration to trigger all token of the user
            $stateId = State::saveState($state, 'privacyidea:privacyidea');
            if (!$this->pi->serviceAccountAvailable()) {
                Logger::error(
                    'privacyIDEA: service account or password is not set in config. Cannot to do trigger challenge.'
                );
            } else {
                $response = null;
                try {
                    $response = $this->pi->triggerChallenge($username);
                } catch (Exception $e) {
                    Utils::handlePrivacyIDEAException($e, $state);
                }

                if (null !== $response) {
                    $stateId = Utils::processPIResponse($stateId, $response);
                }
            }
        } elseif (!empty($this->authProcConfig['tryFirstAuthentication']) && true === $this->authProcConfig['tryFirstAuthentication']) {
            // Call /validate/check with a static pass from the configuration
            // This could already end the authentication with the "passOnNoToken" policy, or it could trigger challenges
            $response = Utils::authenticatePI($state, [
                'otp' => $this->authProcConfig['tryFirstAuthPass'],
            ]);
            if (empty($response->multiChallenge) && $response->value) {
                ProcessingChain::resumeProcessing($state);
            } elseif (!empty($response->multiChallenge)) {
                $stateId = Utils::processPIResponse($stateId, $response);
            }
        }

        $state = State::loadState($stateId, 'privacyidea:privacyidea');

        // This is AuthProcFilter, so step 1 (username+password) is already done. Set the step to 2
        $state['privacyidea:privacyidea:ui']['step'] = 2;
        $stateId = State::saveState($state, 'privacyidea:privacyidea');

        $url = Module::getModuleURL('privacyidea/FormBuilder.php');
        HTTP::redirectTrustedURL($url, [
            'stateId' => $stateId,
        ]);
    }

    /**
     * This function allows to show the debug messages from privacyIDEA server.
     *
     * @param $message
     */
    public function piDebug($message)
    {
        Logger::debug($message);
    }

    /**
     * This function allows to show the debug messages from privacyIDEA server.
     *
     * @param $message
     */
    public function piError($message)
    {
        Logger::error($message);
    }

    /**
     * Check if PrivacyIDEA was disabled by a filter.
     *
     * @param array $state  the global state of simpleSAMLphp
     * @param array $config the config for the PrivacyIDEA server
     *
     * @return bool whether PrivacyIDEA is disabled
     */
    public static function isPrivacyIDEADisabled(array $state, array $config)
    {
        if (isset($config['enabledPath'], $config['enabledKey'])) {
            return isset($state[$config['enabledPath']][$config['enabledKey']][0])
                && !$state[$config['enabledPath']][$config['enabledKey']][0];
        }

        return false;
    }

    /**
     * This function check if user has a token and if not - help to enroll a new one in UI.
     *
     * @param string $stateId
     * @param string $username
     *
     * @return string
     */
    private function enrollToken($stateId, $username)
    {
        assert('string' === gettype($username));
        assert('string' === gettype($stateId));

        $state = State::loadState($stateId, 'privacyidea:privacyidea');

        // Error if no serviceAccount or servicePass
        if (false === $this->pi->serviceAccountAvailable()) {
            Logger::error('privacyIDEA: service account for token enrollment is not set!');
        } else {
            $genkey = 1;
            $type = $this->authProcConfig['tokenType'];
            $description = 'Enrolled with simpleSAMLphp';

            $response = $this->pi->enrollToken($username, $genkey, $type, $description);

            if (!empty($response->errorMessage)) {
                Logger::error(
                    'privacyIDEA: Error code: ' . $response->errorCode . ', Error message: ' . $response->errorMessage
                );
                $state['privacyidea:privacyidea']['errorCode'] = $response->errorCode;
                $state['privacyidea:privacyidea']['errorMessage'] = $response->errorMessage;
            }

            // If we have a response from PI - save QR Code into state to show it soon
            // and enroll a new token for the user
            if (!empty($response->detail->googleurl->img)) {
                $state['privacyidea:tokenEnrollment']['tokenQR'] = $response->detail->googleurl->img;
            }

            return State::saveState($state, 'privacyidea:privacyidea');
        }

        return '';
    }

    /**
     * This is the help function to exclude some IP from 2FA. Only if is set in config.
     *
     * @param $clientIP
     * @param $excludeClientIPs
     *
     * @return bool|void
     */
    private function matchIP($clientIP, $excludeClientIPs)
    {
        assert('string' === gettype($clientIP));
        $clientIP = ip2long($clientIP);

        $match = false;
        foreach ($excludeClientIPs as $ipAddress) {
            if (strpos($ipAddress, '-')) {
                $range = explode('-', $ipAddress);
                $startIP = ip2long($range[0]);
                $endIP = ip2long($range[1]);
                $match = $clientIP >= $startIP && $clientIP <= $endIP;
            } else {
                $match = $clientIP === ip2long($ipAddress);
            }
            if ($match) {
                break;
            }
        }

        return $match;
    }

    /**
     * This function allows the selective deactivation of privacyIDEA for a list of regular expressions which match SAML
     * service provider entityIDs. The filter checks the entityID in the SAML request against a list of regular
     * expressions and sets the state variable $state[enabledPath][enabledKey][0] to false on match, which can be used
     * to disable privacyIDEA. For any value in excludeEntityIDs, the config parameter includeAttributes may be used to
     * enable privacyIDEA for a subset of users which have these attribute values (e.g. memberOf).
     *
     * @param array  $authProcConfig
     * @param string $stateId
     *
     * @return string
     */
    private function checkEntityID($authProcConfig, $stateId)
    {
        Logger::debug('Checking requesting entity ID for privacyIDEA');
        $state = State::loadState($stateId, 'privacyidea:privacyidea');

        $excludeEntityIDs = $authProcConfig['excludeEntityIDs'] ?: [];
        $includeAttributes = $authProcConfig['includeAttributes'] ?: [];
        $setPath = $authProcConfig['setPath'] ?: '';
        $setKey = $authProcConfig['setKey'] ?: '';

        // the default return value is true, privacyIDEA should be enabled by default.
        $ret = true;
        $requestEntityID = $state['Destination']['entityid'];

        // if the requesting entityID matches the given list set the return parameter to false
        Logger::debug('privacyidea:checkEntityID: Requesting entityID is ' . $requestEntityID);
        $matchedEntityIDs = $this->strMatchesRegArr($requestEntityID, $excludeEntityIDs);
        if ($matchedEntityIDs) {
            $ret = false;
            $entityIDKey = $matchedEntityIDs[0];
            Logger::debug('privacyidea:checkEntityID: Matched entityID is ' . $entityIDKey);

            // if there is also a match for any attribute value in the includeAttributes
            // fall back to the default return value: true
            if (isset($includeAttributes[$entityIDKey])) {
                foreach ($includeAttributes[$entityIDKey] as $attrKey => $attrRegExpArr) {
                    if (isset($state['Attributes'][$attrKey])) {
                        foreach ($state['Attributes'][$attrKey] as $attrVal) {
                            $matchedAttrs = $this->strMatchesRegArr($attrVal, $attrRegExpArr);

                            if (!empty($matchedAttrs)) {
                                $ret = true;
                                Logger::debug('privacyidea:checkEntityID: Requesting entityID in ' .
                                                         'list, but excluded by at least one attribute regexp "' . $attrKey .
                                                         '" = "' . $matchedAttrs[0] . '".');
                                break;
                            }
                        }
                    } else {
                        Logger::debug('privacyidea:checkEntityID: attribute key ' .
                                                 $attrKey . ' not contained in request');
                    }
                }
            }
        } else {
            Logger::debug('privacyidea:checkEntityID: Requesting entityID ' .
                                     $requestEntityID . ' not matched by any regexp.');
        }

        $state[$setPath][$setKey][0] = $ret;

        $stateId = State::saveState($state, 'privacyidea:privacyidea');

        if ($ret) {
            $retStr = 'true';
        } else {
            $retStr = 'false';
        }
        Logger::debug('Setting $state[' . $setPath . '][' . $setKey . '][0] = ' . $retStr . '.');

        return $stateId;
    }

    /**
     * This is the help function for checkEntityID() and checks a given string against an array with regular
     * expressions. It will return an array with matches.
     *
     * @param string $str
     *
     * @return array
     */
    private function strMatchesRegArr($str, array $reg_arr)
    {
        $retArr = [];

        foreach ($reg_arr as $reg) {
            if ('/' !== $reg[0]) {
                $reg = '/' . $reg . '/';
            }
            Logger::debug('privacyidea:checkEntityID: test regexp ' . $reg . ' against the string ' . $str);

            if (preg_match($reg, $str)) {
                $retArr[] = $reg;
            }
        }

        return $retArr;
    }
}
